module Estapir

export resources, RESTAPIPlug

using Bukdu
using StructDatabaseMapping
import JSON
import JSON: print
const INVALID_REQUEST = "Invalid Request"

import StructDatabaseMapping: isprimarykey
abstract type AbstractRESTAPIContoller <: ApplicationController end

struct RESTAPIController <: AbstractRESTAPIContoller
    conn::Conn
end


struct InvalidRequest
    invalid_fields::Array
    missing_fields::Array
end
function InvalidRequest(;invalid::Array=[], missing::Array=[])
    return InvalidRequest(invalid, missing)
end

function fieldlist(mapper::DBMapper, ::Type{T}) where T<:Model
    table = mapper.tables[T]
    return [field.struct_field for field in StructDatabaseMapping.fieldlist(table) if !isprimarykey(field, table)]
end

function to_response(s::InvalidRequest) :: Dict
    d = Dict{Any, Any}("status"=>INVALID_REQUEST)
    if length(s.invalid_fields) > 0
        d["invalid_fields"] = s.invalid_fields
    end
    if length(s.missing_fields) > 0
        d["missing_fields"] = s.missing_fields
    end
    return d
end
print(io::IO, s::InvalidRequest) = print(io, to_response(s))

function check_plug(c::RESTAPIController)
    if !haskey(c.conn.private, :modeltype)
        throw("ModelType not present. Did you forget to use the plug?")
    end
end

function validate(params::Dict{Symbol, Any}, mapper::DBMapper, ::Type{T}) where T<:Model
    param_fields = collect(keys(params))
    defined_fields = fieldlist(mapper, T)
    invalid_fields = setdiff(param_fields, defined_fields)
    missing_fields = setdiff(defined_fields, param_fields)
    if length(missing_fields) > 0 || length(invalid_fields) > 0
        return InvalidRequest(invalid=invalid_fields,
                              missing=missing_fields)
    end
    return nothing    

end

function create(c::RESTAPIController, mapper::DBMapper, ::Type{ModelType}, path::String) where ModelType<:Model
    params = Dict{Symbol, Any}(Symbol(key)=>value for (key, value) in c.conn.params)
    errors = validate(params, mapper, ModelType)
    if errors === nothing        
        model = ModelType(;params...)
        insert!(mapper, model)
        c.conn.request.response.status = 201
        push!(c.conn.request.response.headers, Pair("Location", "$(path)/$(model.id.x)"))
        return Bukdu.render(Bukdu.JSON, model)
    else      
        c.conn.request.response.status = 400  
        return Bukdu.render(Bukdu.JSON, errors)
    end
end

function list(c::RESTAPIController, mapper::DBMapper, ::Type{ModelType}, path::String) where ModelType<:Model
    return Bukdu.render(Bukdu.JSON, select_all(mapper, ModelType))
end

function get(c::RESTAPIController, mapper::DBMapper, ::Type{ModelType}, path::String) where ModelType<:Model
    id = c.conn.params[:identifier]
    elem = select_one(mapper, ModelType, pk=id)
    return Bukdu.render(Bukdu.JSON, elem)
end


function resources(::Type{T}, mapper::DBMapper; endpoint::String="") where T<:Model
    model_name = lowercase(string(T))
    pipe1 = Symbol(model_name)
    plug_name = Symbol(string(T) * "Plug")
    endpoint = "$(endpoint)/$(model_name)"

    wrap_function(f) = (c::RESTAPIController) -> f(c, mapper, T, endpoint)
    
    Routing.add_route(post, endpoint, 
                      RESTAPIController, wrap_function(create),
                      Dict{Symbol,DataType}())
    Routing.add_route(get, endpoint, 
                      RESTAPIController, wrap_function(list), 
                      Dict{Symbol,DataType}())
    Routing.add_route(get, endpoint * "/:identifier", 
                      RESTAPIController, wrap_function(get), 
                      Dict{Symbol,DataType}(:identifier=>String))        
end

end # module
